﻿using Microsoft.Extensions.Configuration;
using Phares.Shared;
using Serilog;
using ShellProgressBar;
using View_by_Distance.Copy.Distinct.Models;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Methods;

namespace View_by_Distance.Copy.Distinct;

public class CopyDistinct
{

    private readonly AppSettings _AppSettings;
    private readonly string _WorkingDirectory;
    private readonly Configuration _Configuration;
    private readonly IsEnvironment _IsEnvironment;
    private readonly IConfigurationRoot _ConfigurationRoot;
    private readonly IReadOnlyDictionary<string, string[]> _FileGroups;
    private readonly Property.Models.Configuration _PropertyConfiguration;

    public CopyDistinct(List<string> args, IsEnvironment isEnvironment, IConfigurationRoot configurationRoot, AppSettings appSettings, string workingDirectory, bool isSilent, IConsole console)
    {
        if (isSilent)
        { }
        if (console is null)
        { }
        _AppSettings = appSettings;
        _IsEnvironment = isEnvironment;
        _WorkingDirectory = workingDirectory;
        _ConfigurationRoot = configurationRoot;
        ILogger? log = Log.ForContext<CopyDistinct>();
        Property.Models.Configuration propertyConfiguration = Property.Models.Binder.Configuration.Get(isEnvironment, configurationRoot);
        Configuration configuration = Models.Binder.Configuration.Get(isEnvironment, configurationRoot, propertyConfiguration);
        _PropertyConfiguration = propertyConfiguration;
        _Configuration = configuration;
        propertyConfiguration.Update();
        log.Information(propertyConfiguration.RootDirectory);
        (bool move, List<string[]> filesCollection, bool anyLenFiles, bool moveBack) = Verify();
        _FileGroups = Shared.Models.Stateless.Methods.IPath.GetKeyValuePairs(propertyConfiguration, appSettings.CopyTo, new string[] { appSettings.ResultDirectoryKey });
        List<string> lines = CopyDistinctFilesInDirectories(log, move, filesCollection, anyLenFiles, moveBack);
        if (lines.Any())
            File.WriteAllLines($"D:/Tmp/Phares/{DateTime.Now.Ticks}.tsv", lines);
        _ = Shared.Models.Stateless.Methods.IPath.DeleteEmptyDirectories(propertyConfiguration.RootDirectory);
    }

    private (bool, List<string[]>, bool, bool) Verify()
    {
        if (_AppSettings is null)
            throw new NullReferenceException(nameof(_AppSettings));
        if (_IsEnvironment is null)
            throw new NullReferenceException(nameof(_IsEnvironment));
        if (_Configuration is null)
            throw new NullReferenceException(nameof(_Configuration));
        if (_ConfigurationRoot is null)
            throw new NullReferenceException(nameof(_ConfigurationRoot));
        if (_WorkingDirectory is null)
            throw new NullReferenceException(nameof(_WorkingDirectory));
        if (_PropertyConfiguration is null)
            throw new NullReferenceException(nameof(_PropertyConfiguration));
        bool moveBack;
        const string fileSearchFilter = "*";
        const string directorySearchFilter = "*";
        string copyTo = Path.GetFullPath(_AppSettings.CopyTo);
        bool move = copyTo == _PropertyConfiguration.RootDirectory;
        List<string[]> filesCollection = Shared.Models.Stateless.Methods.IDirectory.GetFilesCollection(_PropertyConfiguration.RootDirectory, directorySearchFilter, fileSearchFilter);
        bool anyLenFiles = filesCollection.Any(l => l.Any(m => m.EndsWith("len")));
        if (!move)
            moveBack = false;
        else
        {
            string directory = Path.Combine(_PropertyConfiguration.RootDirectory, _AppSettings.ResultDirectoryKey);
            if (!anyLenFiles)
                throw new NotSupportedException("Use Mirror-Length app first!");
            if (string.IsNullOrEmpty(_AppSettings.ResultDirectoryKey))
                throw new NotSupportedException("Change appsettings!");
            if (!Directory.Exists(directory))
                moveBack = false;
            else
            {
                if (!Directory.GetFiles(directory, "*", SearchOption.AllDirectories).Any())
                    moveBack = false;
                else
                {
                    move = false;
                    moveBack = true;
                }
            }
        }
        if (move && _AppSettings.IfCanUseId)
            throw new NotSupportedException("Not allowed because it would irreversible!");
        return (move, filesCollection, anyLenFiles, moveBack);
    }

    private static (string[], List<(FileHolder, string)>) GetMoveBackToDoCollection(List<string[]> filesCollection)
    {
        List<(FileHolder, string)> results = new();
        string key;
        string? value;
        string fileName;
        string? directory;
        string destinationFile;
        List<string> distinctFound = new();
        List<string> distinctNeeded = new();
        List<string> distinctDirectories = new();
        Dictionary<string, string> nameToPath = new();
        for (int i = 1; i < 3; i++)
        {
            foreach (string[] files in filesCollection)
            {
                foreach (string file in files.Reverse())
                {
                    fileName = Path.GetFileName(file);
                    if (fileName.EndsWith("len"))
                    {
                        key = fileName[..^3];
                        destinationFile = file[..^3];
                        if (nameToPath.ContainsKey(key))
                            continue;
                        nameToPath.Add(key, destinationFile);
                    }
                    else
                    {
                        if (!distinctNeeded.Contains(file))
                            distinctNeeded.Add(file);
                        if (!nameToPath.ContainsKey(fileName))
                            continue;
                        if (distinctFound.Contains(file))
                            continue;
                        distinctFound.Add(file);
                    }
                }
            }
        }
        foreach (string[] files in filesCollection)
        {
            foreach (string file in files)
            {
                // if (distinctNeeded.Count != distinctFound.Count)
                //     continue;
                fileName = Path.GetFileName(file);
                if (fileName.EndsWith("len"))
                    continue;
                if (!nameToPath.TryGetValue(fileName, out value))
                    continue;
                directory = Path.GetDirectoryName(value);
                if (string.IsNullOrEmpty(directory))
                    continue;
                results.Add(new(new(file), value));
                if (!distinctDirectories.Contains(directory))
                    distinctDirectories.Add(directory);
            }
        }
        return (distinctDirectories.ToArray(), results);
    }

    private List<string> CopyDistinctFilesInDirectories(ILogger log, bool move, List<string[]> filesCollection, bool anyLenFiles, bool moveBack)
    {
        List<string> results = new();
        ProgressBar progressBar;
        string[] distinctDirectories;
        ConsoleKey? consoleKey = null;
        string message = nameof(CopyDistinct);
        List<(FileHolder, string)> toDoCollection;
        int count = filesCollection.Select(l => l.Length).Sum();
        ProgressBarOptions options = new() { ProgressCharacter = '─', ProgressBarOnBottom = true, DisableBottomPercentage = true };
        if (moveBack)
        {
            if (!anyLenFiles)
                throw new NotSupportedException();
            (distinctDirectories, toDoCollection) = GetMoveBackToDoCollection(filesCollection);
        }
        else
        {
            progressBar = new(count, message, options);
            string key = string.IsNullOrEmpty(_AppSettings.ResultDirectoryKey) ? _PropertyConfiguration.ResultAllInOne : _AppSettings.ResultDirectoryKey;
            string[] directories = _FileGroups[key];
            (distinctDirectories, toDoCollection) = Shared.Models.Stateless.Methods.IDirectory.GetToDoCollection(_PropertyConfiguration, _AppSettings.CopyDuplicates, _AppSettings.IfCanUseId, filesCollection, directories, () => progressBar.Tick());
            progressBar.Dispose();
        }
        foreach (string distinctDirectory in distinctDirectories)
        {
            if (!Directory.Exists(distinctDirectory))
                _ = Directory.CreateDirectory(distinctDirectory);
        }
        if (move)
            log.Information($"Ready to Move {toDoCollection.Count} file(s)?");
        else if (!moveBack)
            log.Information($"Ready to Copy {toDoCollection.Count} file(s)?");
        else
            log.Information($"Ready to Move back {toDoCollection.Count} file(s)?");
        for (int y = 0; y < int.MaxValue; y++)
        {
            if (move)
                log.Information("Press \"Y\" key to move file(s), \"N\" key to log file(s) or close console to not move files");
            else if (!moveBack)
                log.Information("Press \"Y\" key to copy file(s), \"N\" key to log file(s) or close console to not copy files");
            else
                log.Information("Press \"Y\" key to move back file(s), \"N\" key to log file(s) or close console to not move back files");
            consoleKey = System.Console.ReadKey().Key;
            if (consoleKey is ConsoleKey.Y or ConsoleKey.N)
                break;
        }
        log.Information(". . .");
        if (consoleKey is null || consoleKey.Value != ConsoleKey.Y)
        {
            if (move || moveBack)
                log.Information("Nothing moved!");
            else
                log.Information("Nothing copied!");
        }
        else
        {
            progressBar = new(count, message, options);
            results.AddRange(Shared.Models.Stateless.Methods.IDirectory.CopyOrMove(toDoCollection, move, moveBack, () => progressBar.Tick()));
            progressBar.Dispose();
            if (move || moveBack)
                log.Information("Done moving");
            else
                log.Information("Done copying");
        }
        return results;
    }

}